package atomicstryker.multimine.common;

import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import java.util.PriorityQueue;
import java.util.Set;

import com.google.common.collect.Lists;
import com.google.common.collect.Maps;

import atomicstryker.multimine.common.network.PartialBlockPacket;
import atomicstryker.multimine.common.network.PartialBlockRemovalPacket;
import net.minecraft.block.Block;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.player.EntityPlayerMP;
import net.minecraft.item.ItemStack;
import net.minecraft.server.MinecraftServer;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.BlockPos;
import net.minecraftforge.common.ForgeHooks;
import net.minecraftforge.common.config.Configuration;
import net.minecraftforge.common.config.Property;
import net.minecraftforge.fml.common.FMLCommonHandler;
import net.minecraftforge.fml.common.eventhandler.SubscribeEvent;
import net.minecraftforge.fml.common.gameevent.PlayerEvent.PlayerLoggedInEvent;
import net.minecraftforge.fml.common.gameevent.TickEvent;
import net.minecraftforge.fml.common.gameevent.TickEvent.Phase;
import net.minecraftforge.fml.common.network.NetworkRegistry.TargetPoint;
import net.minecraftforge.fml.common.registry.GameData;

public class MultiMineServer
{
    private static MultiMineServer instance;
    private static MinecraftServer serverInstance;
    private final HashMap<Integer, List<PartiallyMinedBlock>> partiallyMinedBlocksListByDimension;
    private final BlockRegenQueue blockRegenQueue;
    public final HashMap<String, Boolean> blacklistedBlocksAndTools;

    /**
     * Server instance of Multi Mine Mod. Keeps track of Players having the Mod
     * installed, the Blocks they damage, and the Block regeneration Queue,
     * which is watched by the integrated Tick Handler.
     */
    public MultiMineServer()
    {
        partiallyMinedBlocksListByDimension = Maps.<Integer, List<PartiallyMinedBlock>>newHashMap();
        instance = this;
        blockRegenQueue = new BlockRegenQueue(30, new BlockAgeComparator());
        blacklistedBlocksAndTools = Maps.<String, Boolean>newHashMap();
        
        final Set<Entry<String, Property>> setblocks = MultiMine.instance().config.getCategory("bannedblocks").entrySet();
        for (Entry<String, Property> entry : setblocks)
        {
            blacklistedBlocksAndTools.put(entry.getKey(), entry.getValue().getBoolean(false));
        }
        final Set<Entry<String, Property>> setitems = MultiMine.instance().config.getCategory("banneditems").entrySet();
        for (Entry<String, Property> entry : setitems)
        {
            blacklistedBlocksAndTools.put(entry.getKey(), entry.getValue().getBoolean(false));
        }

        FMLCommonHandler.instance().bus().register(this);
    }

    public static MultiMineServer instance()
    {
        return instance;
    }

    /**
     * Called when a client has a block progression update. Update the
     * already existing partial Block or create one if not present,
     * and send that information back to all interested players.
     * 
     * @param player
     *            player doing the digging
     * @param x
     *            coordinate of Block
     * @param y
     *            coordinate of Block
     * @param z
     *            coordinate of Block
     * @param value
     *            block progression the client reported
     */
    public void onClientSentPartialBlockPacket(EntityPlayerMP player, int x, int y, int z, float value)
    {
        serverInstance = FMLCommonHandler.instance().getMinecraftServerInstance();
        int dimension = player.field_71093_bK;
        MultiMine.instance().debugPrint("multi mine client " + player + " sent progress packet: " + value);
        
        final BlockPos pos = new BlockPos(x, y, z);
        final IBlockState iblockstate = player.field_70170_p.func_180495_p(pos);
        final Block block = iblockstate.func_177230_c();
        if (isItemBanned(player.func_71045_bC()) || isBlockBanned(block, block.func_176201_c(iblockstate)))
        {
            return;
        }
        
        List<PartiallyMinedBlock> partiallyMinedBlocks = getPartiallyMinedBlocksForDimension(dimension);

        if (partiallyMinedBlocks == null)
        {
            partiallyMinedBlocks = Lists.<PartiallyMinedBlock> newArrayList();
            partiallyMinedBlocksListByDimension.put(dimension, partiallyMinedBlocks);
        }

        final PartiallyMinedBlock newblock = new PartiallyMinedBlock(x, y, z, dimension, 0f);
        newblock.setLastTimeMined(System.currentTimeMillis() + MultiMine.instance().getInitialBlockRegenDelay());
        for (PartiallyMinedBlock iterBlock : partiallyMinedBlocks)
        {
            if (iterBlock.equals(newblock))
            {
                iterBlock.setProgress(Math.max(iterBlock.getProgress(), value));
                iterBlock.setLastTimeMined(System.currentTimeMillis() + MultiMine.instance().getInitialBlockRegenDelay());
                MultiMine.instance().debugPrint("Server updating partial block at: ["+x+"|"+y+"|"+z+"], progress now: "+iterBlock.getProgress());

                // send the newly advanced partialblock to all relevant players
                sendPartiallyMinedBlockUpdateToAllPlayers(iterBlock);

                
                if (iterBlock.isFinished() && !block.isAir(player.field_70170_p, pos))
                {
                    MultiMine.instance().debugPrint("Server finishing partial block at: ["+x+"|"+y+"|"+z+"]");
                    // and if its done, destroy the world block
                    player.field_70170_p.func_175715_c(player.func_145782_y(), pos, -1);
                    
                    final int event =
                            ForgeHooks.onBlockBreakEvent(player.field_70170_p, player.field_71134_c.func_73081_b(), player, pos);
                    if (event != -1)
                    {
                        TileEntity tileentity = player.field_70170_p.func_175625_s(pos);

                        ItemStack stack = player.func_71045_bC();
                        if (stack != null && stack.func_77973_b().onBlockStartBreak(stack, pos, player)) return;

                        player.field_70170_p.func_180498_a(player, 2001, pos, Block.func_176210_f(iblockstate));

                        ItemStack itemstack = player.func_71045_bC();
                        boolean canHarvest = iblockstate.func_177230_c().canHarvestBlock(player.field_70170_p, pos, player);

                        if (itemstack != null)
                        {
                            itemstack.func_179548_a(player.field_70170_p, iblockstate.func_177230_c(), pos, player);
                            if (itemstack.field_77994_a == 0)
                            {
                                player.func_71028_bD();
                            }
                        }
                        
                        iblockstate.func_177230_c().func_176208_a(player.field_70170_p, pos, iblockstate, player);
                        boolean removed = iblockstate.func_177230_c().removedByPlayer(player.field_70170_p, pos, player, canHarvest);
                        if (removed)
                        {
                            iblockstate.func_177230_c().func_176206_d(player.field_70170_p, pos, iblockstate);
                        }
                        
                        if (removed && canHarvest)
                        {
                            iblockstate.func_177230_c().func_180657_a(player.field_70170_p, player, pos, iblockstate, tileentity);
                        }

                        // Drop experiance
                        if (removed && event > 0)
                        {
                            iblockstate.func_177230_c().func_180637_b(player.field_70170_p, pos, event);
                        }
                        
                        partiallyMinedBlocks.remove(iterBlock);
                        blockRegenQueue.remove(iterBlock);
                    }
                }
                else
                {
                    blockRegenQueue.offer(iterBlock);
                }
                
                return;
            }
        }

        // else send the new partialblock to all relevant players
        // System.out.println("Server creating new partial block at: ["+x+"|"+y+"|"+z+"]");

        if (partiallyMinedBlocks.size() > 29)
        {
            PartiallyMinedBlock old = partiallyMinedBlocks.get(0);
            sendPartiallyMinedBlockDeleteCommandToAllPlayers(old);
            partiallyMinedBlocks.remove(old);
            blockRegenQueue.remove(old);
        }

        partiallyMinedBlocks.add(newblock);
        blockRegenQueue.offer(newblock);
        sendPartiallyMinedBlockUpdateToAllPlayers(newblock);
    }

    private boolean isBlockBanned(Block block, int meta)
    {
        final String ident = GameData.getBlockRegistry().func_177774_c(block).toString()+"-"+meta;
        Boolean result = blacklistedBlocksAndTools.get(ident);
        if (result != null)
        {
            return result;
        }
        
        final Configuration config = MultiMine.instance().config;
        config.load();
        result = config.get("bannedblocks", ident, false).getBoolean(false);
        config.save();
        blacklistedBlocksAndTools.put(ident, result);
        
        return result;
    }

    private boolean isItemBanned(ItemStack item)
    {
        if (item == null || item.func_77973_b() == null)
        {
            return false;
        }
        
        final String ident = GameData.getItemRegistry().func_177774_c(item.func_77973_b()).toString()+"-"+item.func_77952_i();
        Boolean result = blacklistedBlocksAndTools.get(ident);
        if (result != null)
        {
            return result;
        }
        
        final Configuration config = MultiMine.instance().config;
        config.load();
        result = config.get("banneditems", ident, false).getBoolean(false);
        config.save();
        blacklistedBlocksAndTools.put(ident, result);
        
        return result;
    }

    /**
     * Tells all clients to delete this partially mined Block off their local
     * storage, the server exceeds the max amount of concurrent partial Blocks.
     * 
     * @param block
     *            partial Block to be deleted
     */
    private void sendPartiallyMinedBlockDeleteCommandToAllPlayers(PartiallyMinedBlock block)
    {
        MultiMine.instance().networkHelper.sendPacketToAllAroundPoint(new PartialBlockRemovalPacket(block.getPos()),
                new TargetPoint(block.getDimension(), block.getPos().func_177958_n(), block.getPos().func_177956_o(), block.getPos().func_177952_p(), 30D));
    }

    @SubscribeEvent
    public void onPlayerLogin(PlayerLoggedInEvent event)
    {
        final EntityPlayerMP player = (EntityPlayerMP) event.player;
        int dimension = player.field_70170_p.field_73011_w.func_177502_q();
        final List<PartiallyMinedBlock> partiallyMinedBlocks = getPartiallyMinedBlocksForDimension(dimension);
        if (partiallyMinedBlocks != null)
        {
            for (PartiallyMinedBlock block : partiallyMinedBlocks)
            {
                sendPartiallyMinedBlockToPlayer(player, block);
            }
        }
    }

    /**
     * Helper method to get the correct partial Block list for a World Dimension
     * 
     * @param dim
     *            Dimension of the world
     * @return the List of partial Blocks for the dimension. Can be null, can be
     *         empty.
     */
    private List<PartiallyMinedBlock> getPartiallyMinedBlocksForDimension(int dim)
    {
        return partiallyMinedBlocksListByDimension.get(dim);
    }

    /**
     * Sends a partial Block Packet to all players in the matching dimension and
     * a certain area. Overwrites their local partial Block instances with
     * whatever you send.
     * 
     * @param block
     *            PartiallyMinedBlock instance
     */
    private void sendPartiallyMinedBlockUpdateToAllPlayers(PartiallyMinedBlock block)
    {
        MultiMine.instance().networkHelper.sendPacketToAllAroundPoint(new PartialBlockPacket("server", block.getPos().func_177958_n(), block.getPos().func_177956_o(), block.getPos().func_177952_p(),
                block.getProgress()), new TargetPoint(block.getDimension(), block.getPos().func_177958_n(), block.getPos().func_177956_o(), block.getPos().func_177952_p(), 32D));
    }

    /**
     * Sends a partial Block Packet to a particular player.
     * 
     * @param p
     *            Player targeted
     * @param block
     *            PartiallyMinedBlock instance
     */
    private void sendPartiallyMinedBlockToPlayer(EntityPlayerMP p, PartiallyMinedBlock block)
    {
        MultiMine.instance().networkHelper.sendPacketToPlayer(
                new PartialBlockPacket("server", block.getPos().func_177958_n(), block.getPos().func_177956_o(), block.getPos().func_177952_p(), block.getProgress()), p);
    }

    /**
     * Tick Handler to achieve Block Regeneration. We keep track of our Block
     * age using a PriorityQueue and start repairing Blocks if they get too old.
     */
    @SubscribeEvent
    public void onTick(TickEvent.WorldTickEvent tick)
    {
        if (tick.phase == Phase.END)
        {
            if (blockRegenQueue.isEmpty())
            {
                return;
            }

            PartiallyMinedBlock block;
            for (Iterator<PartiallyMinedBlock> iter = blockRegenQueue.iterator(); iter.hasNext();)
            {
                block = iter.next();
                if (isBlockGone(block))
                {
                    sendPartiallyMinedBlockDeleteCommandToAllPlayers(block);
                    getPartiallyMinedBlocksForDimension(block.getDimension()).remove(block);
                    iter.remove();
                }
            }

            if (blockRegenQueue.isEmpty() || !MultiMine.instance().getBlockRegenEnabled())
            {
                return;
            }

            long curTime = System.currentTimeMillis();
            if (blockRegenQueue.peek().getLastTimeMined() + MultiMine.instance().getBlockRegenInterval() < curTime)
            {
                block = blockRegenQueue.poll();

                block.setProgress(block.getProgress() - 0.1f);
                block.setLastTimeMined(curTime);
                if (block.getProgress() < 0f)
                {
                    // tell everyone to stop tracking this one
                    sendPartiallyMinedBlockDeleteCommandToAllPlayers(block);
                    getPartiallyMinedBlocksForDimension(block.getDimension()).remove(block);
                }
                else
                {
                    // send update about this one to all
                    sendPartiallyMinedBlockUpdateToAllPlayers(block);
                    blockRegenQueue.add(block);
                }
            }
        }
    }

    /**
     * Helper method to determine if a Block was removed by other means
     * (Explosion, Sand/Gravel falling, Pistons...)
     * 
     * @param block
     *            PartiallyMinedBlock to check
     * @return true if the PartiallyMinedBlock Block coordinates return 0 in a
     *         getBlockId check, false otherwise
     */
    private boolean isBlockGone(PartiallyMinedBlock block)
    {
        return serverInstance.func_71218_a(block.getDimension()).func_175623_d(block.getPos());
    }

    /**
     * PriorityQueue sorting Blocks by age, in order to only check one each tick
     */
    private class BlockRegenQueue extends PriorityQueue<PartiallyMinedBlock>
    {
        private static final long serialVersionUID = 1L;

        public BlockRegenQueue(int initialSize, Comparator<PartiallyMinedBlock> comparator)
        {
            super(initialSize, comparator);
        }

        /**
         * Overriding the offer method in order to delete older instances in
         * favor of more recently mined ones.
         */
        @Override
        public boolean offer(PartiallyMinedBlock block)
        {
            if (contains(block))
            {
                this.remove(block);
            }
            return super.offer(block);
        }
    }

    /**
     * Comparator to help sort PartiallyMinedBlock instances by their age
     */
    private class BlockAgeComparator implements Comparator<PartiallyMinedBlock>
    {
        @Override
        public int compare(PartiallyMinedBlock b1, PartiallyMinedBlock b2)
        {
            if (b1.getLastTimeMined() < b2.getLastTimeMined())
            {
                return -1;
            }
            return 1;
        }
    }
}
